package lsedit;

import java.awt.Component;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.event.*;

import java.util.*;
import java.io.*;
import javax.swing.*;

// Entities in a landscape 

public class BaseEntity extends EntityInstance /* extends LandscapeObject extends JComponent */ implements MouseListener, MouseMotionListener, DiagramCoordinates
{
	protected static EntityInstance m_currentDescEntity = null;

	protected static boolean has3Dlook = true;

	private   Cardinal[] m_dstCardinals;

	public static void set3Dlook(boolean state) 
	{
		has3Dlook = state;
	}

	protected BaseEntity(Diagram dg) 
	{
		super(dg);
	}

	public BaseEntity(EntityClass parentClass, String id, Diagram dg) 
	{
		super(parentClass, id, dg);

		switch(getStyle()) {
		case EntityClass.ENTITY_STYLE_GROUP:
		case EntityClass.ENTITY_STYLE_LABELLED_GROUP:
			setSize(getWidth(), getWidth());
		}
		addMouseListener(this);
		addMouseMotionListener(this);
	}

	protected void clearGroupFlag() 
	{
		if (isMarked(GROUP_MARK|GROUPKEY_MARK)) {
			nandMark(GROUP_MARK|GROUPKEY_MARK);
			repaint();
		}
	}

	protected void setGraphicsColor(Graphics g, Color explicit) 
	{
		if (explicit != null) {
			g.setColor(explicit);
		} else if (isMarked(REDBOX_MARK) && !isOpen()) {
			g.setColor(Color.red.darker());
		} else {
			g.setColor(getBackground());
		}
	}

	public void addToDiagram(JComponent under)
	{
//		System.out.println("BaseEntity:addToDiagram " + this + " " + isOpen() + " " + numChildren());
		removeAll();
		orMark(DIAGRAM_MARK);

		if (isOpen()) {
			Enumeration		en;
			EntityInstance	e;
			for (en = getChildren(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.addToDiagram(this); // Add all children of open parent to diagram
		}	}
//		System.out.println("BaseEntity.addToDiagram(" + this + ")");
		under.add(this);		// Add this entity to diagram
		// Adding causes call to validate
	}
	
	// Reset all cardinals associated with this entity and things it contains

	public void resetCardinals(int numRelations) 
	{
		Enumeration en;

		if (m_dstCardinals == null) {
			if (numRelations > 0) {
				m_dstCardinals = new Cardinal[numRelations];
			}
		} else {
			int		 i;
			Cardinal cardinal;
			
			for (i = 0; i < numRelations; ++i) {
				cardinal = m_dstCardinals[i];
				if (cardinal != null) {
					cardinal.reset();
		}	}	}

		for (en = getChildren(); en.hasMoreElements(); ) {
			BaseEntity e = (BaseEntity) en.nextElement();
			e.resetCardinals(numRelations);
		}
	}

	public void incCardinal(int ind) 
	{
		Cardinal cardinal;
		
		cardinal = m_dstCardinals[ind];
		if (cardinal == null) {
			cardinal            = new Cardinal();
			m_dstCardinals[ind] = cardinal;
		}
		cardinal.inc();
	}

	// Calculate the number of edges into me

	public void calcEdgeCardinals(EntityInstance closedEntity, boolean hOnly)
	{
		Enumeration			en;
		RelationInstance	ri;
		EntityInstance		src, dst, target;
		BaseEntity			e;
		int					ind;

//		System.out.println("EntityInstance.calcEdgeCardinals for " + this);

		// For every edge that comes to us (something in or beneath entities in the diagram)
		for (en = m_dstRelList.elements(); en.hasMoreElements(); ) {
			ri  = (RelationInstance) en.nextElement();
		
			// The destination of ri comes to our entity

			if (closedEntity != null) {
				src = ri.m_src;
				if (closedEntity.hasDescendant(src)) {
					// Ignore internal loops beneath a closed entity
					continue;
				}
				target = closedEntity;
			} else {
				target = this;
			}

			if (!hOnly || ri.getHighlightFlag()) {
				ind = ri.getRelationClass().getNid();
				((BaseEntity) target).incCardinal(ind);
			}
		}

		if (closedEntity == null) {
			if (!isOpen()) {
				closedEntity = this;
		}	}

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (BaseEntity) en.nextElement();
			e.calcEdgeCardinals(closedEntity, hOnly);
	}	}
	
	public void showCardinals()
	{
		Enumeration	en;
		BaseEntity	e;
		int			width, height, cnt;

		width    = getWidth();
		height   = getHeight();

		if (m_dstCardinals != null && width > 0 && height > 0) {
			Cardinal	cardinal;
			int			i, diagramX, top, w, h, w1;
			FontMetrics fm;
			RelationClass rc;
			double		f;

			fm       = getFontMetrics(smallFont);
			h        = fm.getAscent();
			diagramX = getDiagramX();
			top      = getDiagramY() + height + MARGIN;

			for (i = m_dstCardinals.length; i > 0; ) {
				cardinal = m_dstCardinals[--i];
				if (cardinal != null) {
					cnt      = cardinal.getCnt();
					if (cnt > 0) {
						rc = dg.numToRelationClass(i);
						f  = rc.getIOfactor();
						w  = fm.stringWidth("" + cnt);
						cardinal.setBackground(rc.getObjectColor());
						cardinal.setTextSize(w, h);
						cardinal.setCenterTop(diagramX + ((int) (((double) width)*f)), top);
						dg.add(cardinal /* JLayeredPane.PALETTE_LAYER */);		// Add this cardinal to the diagram
		}	}	}	}

		for (en = getChildren(); en.hasMoreElements(); ) {
			e = (BaseEntity) en.nextElement();
			e.showCardinals();
	}	}

	public void moveCardinals()
	{
		Enumeration	en;
		BaseEntity	e;
		int			width;

		// Move cardinals in sync with entity

		width = getWidth();
		if (m_dstCardinals != null) {
			Cardinal		cardinal;
			int			i, diagramX, top;
			RelationClass	rc;
			double			f;

			diagramX = getDiagramX();
			top      = getDiagramY() + getHeight() + MARGIN;
			for (i = m_dstCardinals.length; i > 0; ) {
				cardinal = m_dstCardinals[--i];
				if (cardinal != null) {
					rc       = dg.numToRelationClass(i);
					f        = rc.getIOfactor();
					cardinal.setCenterTop(diagramX + ((int) (((double) width)*f)), top);
					cardinal.revalidate();
		}	}	}
	}

	public void paintShape(Graphics g, int x, int y, int width, int height, boolean fill, Color fillColor) 
	{
		if (width > 0 && height > 0) {
			switch (getStyle()) {
				case EntityClass.ENTITY_STYLE_3DBOX:
				{
					if (fill) {
						setGraphicsColor(g, fillColor);
						g.fill3DRect(x+1, y+1, width-1, height-1, true);
					}
					setGraphicsColor(g, null);
					g.draw3DRect(x, y, width, height, true);
					break;
				}

				case EntityClass.ENTITY_STYLE_2DBOX:
				{
					if (fill) {
						setGraphicsColor(g, fillColor);
						g.fillRect(x, y, width, height);
					}
					g.setColor(Color.black);
					g.drawRect(x, y, width, height);
					break;
				}

				case EntityClass.ENTITY_STYLE_FILE:
				{
					int fd = Math.min(Math.min(width, height)/2, 8 /* FLAP_DIM */);

					int[] xp = new int[6];
					int[] yp = new int[6];

					xp[0] = x;
					yp[0] = y;
					xp[1] = x + width - fd;
					yp[1] = y;
					xp[2] = x + width - 1;
					yp[2] = y + fd;
					xp[3] = xp[2];
					yp[3] = y + height - 1;
					xp[4] = x;
					yp[4] = yp[3];
					xp[5] = xp[0];
					yp[5] = yp[0];

					if (fill) {
						setGraphicsColor(g, fillColor);
						g.fillPolygon(xp, yp, 6);
					}
					g.setColor(Color.darkGray);
					g.drawPolygon(xp, yp, 6);

					xp[0] = xp[1];
					yp[0] = yp[2];
					xp[3] = xp[0];
					yp[3] = yp[0];

					g.drawPolygon(xp, yp, 4);
					break;
				}

				case EntityClass.ENTITY_STYLE_DISK:
				{
					int ad = Math.min(Math.min(width, height)/2, 8 /* ARC_DIM */);

					// Draw a cylinder

					if (fill) {
						setGraphicsColor(g, fillColor);
						g.fillRect(x, y+ad/2, width, height-ad);
						g.fillOval(x, y, width, ad);
						g.fillOval(x, y+height-ad, width, ad);
					}
					g.setColor(Color.black);

					g.drawOval(x, y, width, ad);
					g.drawArc(x, y+height-ad-1, width, ad, 180, 180);
					g.drawLine(x, y+ad/2, x, y+height-ad/2);
					g.drawLine(x+width-1, y+ad/2, x+width-1, y+height-ad/2);
					break;
				}

				case EntityClass.ENTITY_STYLE_FOLDER:
				{
					// Compute flap size

					int fw = ((int) (((double) width) * .4));
					int fh = Math.max(3 /* MIN_FLAP_HT */, Math.min(8 /* MAX_FLAP_HT */, ((int) (height * .2))));
					int fm = 4 /* FLAP_MARGIN */;

					if (fw < 75 /* MIN_FLAP_WD */) {
					   fw += fw/2;
					}
					if (fw < 10 /* TINY_FLAP_WD */) {
						fw = Math.min(fw + 4 /* FLAP_MARGIN */, width - width/3);
						fm = 0;
					}

					int[] xp = new int[9];
					int[] yp = new int[9];

					xp[0] = x;
					yp[0] = y+fh;
					xp[1] = x+fm;
					yp[1] = y+fh;
					xp[2] = xp[1]+fh/2;
					yp[2] = y;
					xp[3] = x + fw;
					yp[3] = yp[2];
					xp[4] = xp[3]+fh/2;
					yp[4] = yp[0];
					xp[5] = x + width-1;
					yp[5] = yp[0];
					xp[6] = xp[5];
					yp[6] = y + height-1;
					xp[7] = xp[0];
					yp[7] = yp[6];
					xp[8] = xp[0];
					yp[8] = yp[0];

					if (fill) {
						setGraphicsColor(g, fillColor);
						g.fillPolygon(xp, yp, 9);
					}

					g.setColor((isMarked(REDBOX_MARK) && isOpen()) ? Color.red : Color.black);
					g.drawPolygon(xp, yp, 9);

					if (!isOpen()) {
						g.drawLine(xp[1], yp[1], xp[4], yp[4]);
					}
					break;
				}

				case EntityClass.ENTITY_STYLE_SOURCEOBJ:
				{
					if (fill) {
						setGraphicsColor(g, fillColor);
						g.fillOval(x, y, width, height);
					}
					g.setColor((isMarked(REDBOX_MARK) && isOpen()) ? Color.red : Color.black);
					g.drawOval(x, y, width, height);
					break;
				}

				case EntityClass.ENTITY_STYLE_CLASS:
				{
					int arc_w = width/5;
					int arc_h = height/5;
					int arc   = Math.min(arc_w, arc_h);

					if (fill) {
						setGraphicsColor(g, fillColor);
						g.fillRoundRect(x, y, width, height, arc, arc);
					}
					g.setColor(Color.black);
					g.drawRoundRect(x, y, width, height, arc, arc);
					break;
				}

				case EntityClass.ENTITY_STYLE_GROUP:
				{
					Util.drawOutlineBox(g, x, y, width, height, has3Dlook);
					break;
				}

				case EntityClass.ENTITY_STYLE_LABELLED_GROUP:
				{
					g.setFont(smallFont);
					Util.drawGroupBox(g, x, y, width, height, getLabel(), getLabelColour(), has3Dlook);
					break;
				}
/*
				default:
				{
					// For debugging (This shows the root)
					g.setColor(Color.red);
					g.drawLine(x, y,        x+width-1, y+height-1);
					g.drawLine(x, y+height-1, x+width-1, y);
				}
*/
	}	}	}	

	public void paintMap(Graphics g, int x, int y, int width, int height, EntityInstance onPath, int depth)
	{
		if (width > 0 && height > 0) {
			Enumeration		children;
			EntityInstance	child;
			int				x1, y1,width1, height1;
			Color			color;

			if (this == onPath) {
				color = Color.green;
			} else {
				color = getObjectColor();
			} 
			paintShape(g, x, y, width, height, !hasChildren() || depth == 1 /* then fill */, color);

			if (depth < 1) {
				for (children = getChildren(); children.hasMoreElements(); ) {
					child = (EntityInstance) children.nextElement();
					x1      = x + (int) (width  * child.xRelLocal());
					y1      = y + (int) (height * child.yRelLocal()); 
					width1  = (int) (width  * child.widthRelLocal());
					height1 = (int) (height * child.heightRelLocal());
					child.paintMap(g, x1, y1, width1, height1, onPath, depth+1);
	}	}	}	}

	public void paintComponent(Graphics g) 
	{
		int	width  = getWidth();
		int height = getHeight();
		
//		System.out.println("BaseEntity.paintComponent: " + this + " bounds=" + getBounds() + " clip=" + g.getClipBounds() + " style=" + getStyle() + " parentClass=" + m_parentClass);		// IJD
/*		java.lang.Thread.dumpStack();
		System.out.println("-----");
*/

		paintShape(g, 0, 0, width, height, true /* fill */, null /* No given fill color */); 

		if (isMarked(HIGHLIGHT_EDGE_MARK)) {
			g.setColor(Color.white);
			g.drawRect(0, 0, width, height);
			g.drawRect(1, 1, width-1, height-1);
		}

		if (!isOpen() && hasChildren() && !isMarked(CLIENT_MARK|SUPPLIER_MARK)) {

			/*  Draw a small mark as shown in top left of object

				x---
				|
				|   --------
				|  |        |
				   |    |   |
				   |    |   |
	  			   |  ----  |
				   |    |   |
				   |    |   |
				   |        |
				   |________|
			*/

			if (width >= CONTENTS_FLAG_DIM*2 && height >= CONTENTS_FLAG_DIM*2) {
				int x1, y1;
		
				g.setColor(getObjectColor().darker());

				g.drawRect(3, 8, CONTENTS_FLAG_DIM /* 8 */, CONTENTS_FLAG_DIM);

				// g.setColor(Color.black);
				g.drawLine(5, 12, 9, 7);
				g.drawLine(7, 10, 7, 9);
		}	}

		if (isMarked(DRAW_CLIENT_MARK)) {
			// Draw flag that shows client exists
			Util.drawArrow(g, width - 5, (int) Util.ARROW_L, (int) (width - Util.ARROW_L*2 - 5), (int) Util.ARROW_L);
		}

		if (isMarked(DRAW_SUPPLIER_MARK)) {
			Util.drawArrow(g, (int) (width - Util.ARROW_L*2 - 5), (int) (Util.ARROW_L), (int) (width - 5), (int) Util.ARROW_L);
		}

		if (getContainedBy() != null) {
			// Draw our own label

			if (isLabelDrawable()) {
				g.setColor(getLabelColour());
				if (isOpen()) {
					drawTopLeftLabel(g);
				} else {

					// centered, perhaps multi-line
					String label1 = getLabel();
					if (!isMarked(CLIENT_MARK|SUPPLIER_MARK)) {
						g.setFont(closedFont[START_FONT_NUM + m_fontDelta]);
					} else {
						g.setFont(smallFont);
						EntityInstance pe = getEnterableParent();
						if (pe != null) {
							label1 = pe.getLabel() + " .\n" + label1;
					}	}
					Util.drawStringWrapped(g, label1, MARGIN, MARGIN, getWidth()-MARGIN*2, getHeight()-MARGIN*2, true, false);
				}
		}	} 
	
		// Put flags around the edges of the box

		if (getGroupFlag()) {
			g.setColor(getLabelColor());

			int pdim = (height < 20) ? 4 : 6;

			// Draw resize points

			if (getGroupKeyFlag()) {
				g.fillRect(1,			   1,				pdim, pdim);
				g.fillRect(width/2-pdim/2, 1,				pdim, pdim);
				g.fillRect(width-pdim,	   1,				pdim, pdim);
				g.fillRect(1,			   height/2-pdim/2, pdim, pdim);
				g.fillRect(width-pdim,	   height/2-pdim/2, pdim, pdim);
				g.fillRect(1,			   height-pdim,		pdim, pdim);
				g.fillRect(width/2-pdim/2, height-pdim,		pdim, pdim);
				g.fillRect(width-pdim,	   height-pdim,		pdim, pdim);
			}  else {
				// Stupidity here.. An outline is one extra byte wide and high because the two
				// edges of a one pixel box occupy 2 pixels.

				g.drawRect(1,			   1,				pdim, pdim);
				g.drawRect(width/2-pdim/2, 1,				pdim, pdim);
				g.drawRect(width-pdim-1,   1,				pdim, pdim);
				g.drawRect(1,			   height/2-pdim/2, pdim, pdim);
				g.drawRect(width-pdim-1,   height/2-pdim/2, pdim, pdim);
				g.drawRect(1,			   height-pdim-1,	pdim, pdim);
				g.drawRect(width/2-pdim/2, height-pdim-1,	pdim, pdim);
				g.drawRect(width-pdim-1,   height-pdim-1,	pdim, pdim);
			}
		}
	}

	public boolean isTransparent() 
	{
		switch (getStyle()) {
		case EntityClass.ENTITY_STYLE_CLASS:
		case EntityClass.ENTITY_STYLE_LABELLED_GROUP:
			return true;
		default:
			return super.isTransparent();
	}	}

	public boolean isCloseable() 	
	{
		switch (getStyle()) {
		case EntityClass.ENTITY_STYLE_CLASS:
		case EntityClass.ENTITY_STYLE_LABELLED_GROUP:
			return false;
		default:
			return super.isCloseable();
	}	}

	public boolean isLabelDrawable() 	
	{
		switch (getStyle()) {
		case EntityClass.ENTITY_STYLE_CLASS:
		case EntityClass.ENTITY_STYLE_LABELLED_GROUP:
			return false;
		default:
			return super.isLabelDrawable();
	}	}

	public boolean isEnterable() 	
	{
		switch (getStyle()) {
		case EntityClass.ENTITY_STYLE_CLASS:
		case EntityClass.ENTITY_STYLE_LABELLED_GROUP:
			return false;
		default:
			return super.isEnterable();
	}	}

	protected void drawTopLeftLabel(Graphics g) 
	{
		switch (getStyle()) {
		case EntityClass.ENTITY_STYLE_FOLDER:

			g.setFont(smallFont);
			// Compute flap size
			int fw = (int) (((double) getWidth()) * .4);
			int fh = Math.max(3 /* MIN_FLAP_HT */, Math.min(8 /* MAX_FLAP_HT */, ((int) (getHeight() * .2))));

			if (fw < 75 /* MIN_FLAP_WD */) {
			   fw += fw/2;
			}
			if (fw < 10 /* TINY_FLAP_WD */) {
				fw = Math.min(fw + 4 /* FLAP_MARGIN */, getWidth() - getWidth()/3);
			}
			Util.drawStringClipped(g, getLabel(), 4 /* FLAP_MARGIN */+fh/2+2, 0, (double) fw, getHeight()-MARGIN*2);
			break;
		case EntityClass.ENTITY_STYLE_CLASS:
			g.setFont(openFont);
			Util.drawStringClipped(g, getLabel(), MARGIN*3, MARGIN, getWidth()-MARGIN*2, getHeight()-MARGIN*2);
			break;
		default:
			super.drawTopLeftLabel(g);
		}

	}

	// Assuming p1 is in this object and p2 is outside,
	// return the point that is the intersection between this object 
	// and the line segment p1->p2.

	public RealPoint calculateIntercept(RealPoint p1, RealPoint p2) 
	{
		double x = 0.0;
		double y = 0.0;

		double t, b, l, r;	// Top, bottom, left, and right
		double x1, y1, x2, y2;

		t = getY() -1;
		b = getY() + getHeight() + 1;
		l = getX() - 1;
		r = getX() + getWidth() + 1;

		boolean found = true;

		x1 = p1.getX();
		y1 = p1.getY();
		
		if (x1 < l || x1 > r || y1 < t || y1 > b) {
			MsgOut.println("calculateIntercept: p1 is not in box");
			return new RealPoint(x, y);
		}

		x2 = p2.getX();
		y2 = p2.getY();
		
		if (x1 == x2 && y1 == y2) {
			MsgOut.println("calculateIntercept: p1 == p2");
			return new RealPoint(x, y);
		}

		if (r <= x2) {
			x = r;
		} else if (x2 < l) {
			x = l;
		} else {
			found = false;
		}

		if (found) {
			// y = mx + b
			y = y1 + (x - x1) * ( (y2 - y1) / (x2 - x1) );
			found = ( t <= y && y <= b );
		}

		if (!found) {
			found = true;

			if (y2 <= t) {
				y = t;
			} else if (y2 >= b) {
				y = b;
			} else {
				found = false;
			}

			if (found) {
				x = x1 + (y - y1) * ( (x2 - x1) / (y2 - y1) );
				found = (l <= x && x <= r);
			}

		}

		if (!found) {
			MsgOut.println("calculateIntercept failed: " + x1 + " " + y1 + " " + x2 + " " + y2); 
		}
		return new RealPoint(x, y);
	}

	public int overResizeTab(int x1, int y1) 
	{
		int xLeft, xMid, xRight;
		int	yTop,  yMid, yBottom;
		int	pos;

		xLeft  = getDiagramX();

		if (x1 < xLeft) {
			// Outside object
			return RSZ_NONE;
		}
		if (x1 <= xLeft + 6) {
			pos = 0;
		} else {
			xRight = xLeft + getWidth();
			if (x1 > xRight) {
				// Outside object
				return RSZ_NONE;
			}
			if (x1 >= xRight - 6) {
				pos = 2;
			} else {
				xMid   = (xLeft + xRight)  / 2;
				if (x1 < xMid - 3 || x1 > xMid + 3) {
					return RSZ_NONE;
				}
				pos = 1;
		}	}

		yTop = getDiagramY();
		if (y1 < yTop) {
			// Outside object
			return RSZ_NONE;
		}
		if (y1 > yTop + 6) {
			yBottom = (int) (yTop + getHeight());
			if (y1 > yBottom) {
				// Outside object
				return RSZ_NONE;
			}
			if (y1 >= yBottom - 6) {
				pos += 6;
			} else {
				yMid = (yTop  + yBottom) / 2;
				if (y1 < yMid - 3 || y1 > yMid + 3) {
					return RSZ_NONE;
				}
				pos += 3;
		}	}

		switch (pos) {
		case 0:
			return RSZ_NW;
		case 1:
			return RSZ_N;
		case 2:
			return RSZ_NE;
		case 3:
			return RSZ_W;
		case 5:
			return RSZ_E;
		case 6:
			return RSZ_SW;
		case 7:
			return RSZ_S;
		case 8:
			return RSZ_SE;
		}
		return RSZ_NONE;
	}

	public boolean isPointOverIO(EdgePoint pt, int x, int y) 
	{
		int	x1, y1;

		x1 = (int) (pt.getX() + 0.5);	// round
		y1 = (int) (pt.getY() + 0.5);	// round	

		return ((x1 - RelationInstance.NEAR_PIXEL_SIZE/2) < x && x < (x1 + RelationInstance.NEAR_PIXEL_SIZE/2) && (y1 - RelationInstance.NEAR_PIXEL_SIZE/2) < y && y < (y1 + RelationInstance.NEAR_PIXEL_SIZE/2));
	}

	public void drawHighlight() 
	{
		if (!isMarked(HIGHLIGHT_EDGE_MARK)) {
			orMark(HIGHLIGHT_EDGE_MARK);
			repaint();
	}	}

	public void undrawHighlight() 
	{
		if (isMarked(HIGHLIGHT_EDGE_MARK)) {
			nandMark(HIGHLIGHT_EDGE_MARK);
			repaint();
	}	}

	public void drawOutline(Graphics g, Rectangle rect)
	{
		g.drawRect(rect.x, rect.y, rect.width, rect.height);
	}

	private EdgePoint getPoint(EdgePoint[] pts, RelationClass rc, int side) 
	{
		int		  nid;
		EdgePoint cached;

		nid	   = rc.getNid();
		cached = pts[nid];

		// Cache calculated points

		if (cached == null || cached.getRc() != rc) {

			double f  = rc.getIOfactor();
			double wf = 0.0;
			double hf = 0.0;

			switch(side) {
			case EdgePoint.TOP:
				wf = f;					// Put at entity x + width * f 
				hf = 0.0;				// But at top
				break;
			case EdgePoint.BOTTOM:
				wf = f;
				hf = 1.0;				// But at bottom
				break;
			case EdgePoint.LEFT:
				wf = 0.0;				// But at left
				hf = f;
				break;
			default:
				wf = 1.0;				// Put at right
				hf = f;
				break;
			}
			if (cached == null) {
				pts[nid] = cached = new EdgePoint(this, rc, wf, hf);
			} else {
				cached.setRc(rc);
			}
			cached.isDefault = true;
			cached.side		 = side;
			cached.rescale();	// Compute actual point given current position of e
		}
/*
		if (side == EdgePoint.TOP && getLabel().equals("code")) {
			System.out.println("BaseEntity.getPoint " + cached);
		}
*/  
		return cached;
	}

	public EdgePoint getPoint(RelationClass rc, int side) 
	{
		EdgePoint[] cache;

		switch(side) {
		case EdgePoint.TOP:
			cache = topPoints;
			if (cache == null) {
				topPoints = cache = new EdgePoint[dg.numRelationClasses()];
			}
			break;
		case EdgePoint.BOTTOM:
			cache = bottomPoints;
			if (cache == null) {
				bottomPoints = cache = new EdgePoint[dg.numRelationClasses()];
			}
			break;
		case EdgePoint.LEFT:
			cache = leftPoints;
			if (cache == null) {
				leftPoints = cache = new EdgePoint[dg.numRelationClasses()];
			}
			break;
		default:
			cache = rightPoints;
			if (cache == null) {
				rightPoints = cache = new EdgePoint[dg.numRelationClasses()];
		}	 }
		return getPoint(cache, rc, side);
	}


	// At what point does the ri relation hit this, given that it goes from the box specified by srcLyt to dstLyt
	// N.B. Most of the time anyway it appears that srcLyt is the srcLyt of this.

	public EdgePoint getOutPoint(RelationInstance ri, int edge_mode, Rectangle srcLyt, Rectangle dstLyt) 
	{
		RelationClass rc = ri.getRelationClass();
		int nid = rc.getNid();

		if (edge_mode == Do.TB_EDGE_STATE) {
			// Only allow edges to be from top or bottom
			BaseEntity src = (BaseEntity) ri.getSrc();
			if (this == src || src.hasAncestor(this)) {
				return getPoint(rc, EdgePoint.BOTTOM);
			}
			return getPoint(rc, EdgePoint.TOP);
		}
		if ((dstLyt.y - (srcLyt.y+srcLyt.height)) > SEP_THRESHOLD) {

			// dstLyt strictly below srcLyt
			return getPoint(rc, EdgePoint.BOTTOM);
		}

		if ((srcLyt.y - (dstLyt.y+dstLyt.height)) > SEP_THRESHOLD) {
			// srcLyt strictly below dstLyt
			return getPoint(rc, EdgePoint.TOP);
		}

		if ((dstLyt.x - (srcLyt.x+srcLyt.width)) > SEP_THRESHOLD) {
			return getPoint(rc, EdgePoint.RIGHT);
		}

		if ((srcLyt.x - (dstLyt.x+dstLyt.width)) > SEP_THRESHOLD) {
			return getPoint(rc, EdgePoint.LEFT);
		}

		if (dstLyt.y > (srcLyt.y+srcLyt.height)) {
			return getPoint(rc, EdgePoint.BOTTOM);
		}
		return getPoint(rc, EdgePoint.TOP);
	}

	protected EdgePoint testMouseOverIO(EdgePoint[] pts, int x, int y) {

		if (pts == null) {
			return null;
		}

		for (int i = 0; i < pts.length; i++) {
			EdgePoint pt = pts[i];

			if (pt != null) {
				if (isPointOverIO(pt, x, y)) {
					return pt;
				}
			}
		}
		return null;
	}

	public EdgePoint getMouseOverIO(int x, int y) 
	{
		EdgePoint pt = testMouseOverIO(topPoints, x, y);
		if (pt != null) {
			return pt;
		}


		pt = testMouseOverIO(bottomPoints, x, y);
		if (pt != null) {
			return pt;
		}


		pt = testMouseOverIO(leftPoints, x, y);
		if (pt != null) {
			return pt;
		}
		return testMouseOverIO(rightPoints, x, y);
	}

	// MouseListener interface

	public void mouseClicked(MouseEvent e)
	{
	}


	public void mouseEntered(MouseEvent e)
	{
	/*
	 	if (m_label.equals("code")) {
			System.out.println("BaseEntity.mouseEntered code " + getDiagramBounds());
			repaint();
		}
	*/
	}

	public void mouseExited(MouseEvent e)
	{
	}

	public void mousePressed(MouseEvent ev)
	{
//		System.out.println("BaseEntity.mousePressed " + this);

		dg.entityPressed(ev, this, ev.getX() + getDiagramX(), ev.getY() + getDiagramY());
	}

	public void mouseReleased(MouseEvent ev)
	{
		dg.entityReleased(ev, this, ev.getX() + getDiagramX(), ev.getY() + getDiagramY());
	}

	// MouseMotionListener interface

	public void mouseDragged(MouseEvent ev)
	{
		dg.entityDragged(ev, this, ev.getX() + getDiagramX(), ev.getY() + getDiagramY());
	}

	public void mouseMoved(MouseEvent ev)
	{
		if (this != m_infoShown) {
			m_infoShown = this;

			String str = getLabel();
			EntityInstance pe = getEnterableParent();

			if (pe != null) {
				str = pe.getLabel() + " . " + str /* + " " + getDiagramX() + "x" + getDiagramY() + " " + getX() + "X" + getY() */;
			} 
			dg.getLs().showInfo(str);

			if (m_currentDescEntity != this && !isOpen() && getEntityClass() != null) {
				String label = getLabel();
				String title = getTitle(); 
				String desc  = getDescription();

				if (desc.length() == 0) {
					desc = "The " + getLabel() + " " + getEntityClass().getLabel() + ".";
				}

				String topline = " (" + getEntityClass().getLabel() + (hasChildren() ? " - " + numChildren() + " items)" : ")" );

				if (title != null) {
					topline = title + topline;
				} else {
					topline = label + topline;
				}
				dg.getLs().setRightTextBox(topline, desc);
				m_currentDescEntity = this;
		}	}
		dg.movedOverThing(ev, this, ev.getX() + getDiagramX(), ev.getY() + getDiagramY());
	}
	
	// Diagram coordinates interface (no multiple inheritence in Java)
	
	public int getDiagramX()
	{
		return(m_diagramX);
	}
	
	public int getDiagramY()
	{
		return(m_diagramY);
	}

	
}

