package lsedit;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;

import java.net.*;
import java.io.*;
import java.util.*;
import javax.swing.*;

public class LandscapeEditorApp extends LandscapeEditorCore /* extends LandscapeViewerCore  extends JPanel */ implements ToolBarEventHandler 
{
	/* Implements the standalone version of the editor */

	protected JMenu m_optionMenu;

	protected static int openFrames = 1;

	/*
	 * Public methods
	 */

	public JMenuBar genMenu() 
	{ 
		JMenuBar mb = new JMenuBar();
		JMenu	 m, m1;
		MyMenuItem mi;
		MyCheckBoxMenuItem cmi;


		// Build File Menu

		m = new JMenu("File");
		Do.landscapeMenuItem(m, this);
		m.addSeparator();
		Do.printLandscapeMenuItem(m, this);
		m.addSeparator();
		Do.quitMenuItem(m, this);

		mb.add(m);

		m = new JMenu("Edit");

		Do.refreshMenuItem(m, this);

		m1 = new JMenu("Draw");
		Do.cutMenuItem(m1, this);
		Do.pasteMenuItem(m1, this);
		Do.drawMenuItem(m1, this);
		Do.deleteEdgeMenuItem(m1, this);

		Do.editClassMenuItem(m1, this);
		m.add(m1);

		m1 = new JMenu("Find");
		Do.findMenuItem(m1, this);
		m.add(m1);

		mb.add(m);

		m = new JMenu("Arrange");
		m1 = new JMenu("Align");
		Do.alignMenuItem(m1, this);
		Do.gridMenuItem(m1, this);

		m.add(m1);

		m1 = new JMenu("Scale");
		Do.scaleMenuItem(m1, this);
		m.add(m1);

		m1 = new JMenu("Group");
		Do.fitMenuItem(m1, this);
		Do.groupMenuItem(m1, this);
		Do.groupAllMenuItem(m1, this);
		m.add(m1);

		m1 = new JMenu("Label");
		fontMenuItem(m1, this);
		Do.fitLabelMenuItem(m1, this);
		m.add(m1);
	
		mb.add(m);

		m = new JMenu("Layout");
		LandscapeLayouter layout;
		layout = new FlipLayoutHorizontally(this);
		new MyMenuItem(m, layout.getMenuLabel(), layout, 0, 0);
		layout = new FlipLayoutVertically(this);
		new MyMenuItem(m, layout.getMenuLabel(), layout, 0, 0);
		m.addSeparator();
		layout = new GridGroupHorizontally(this);
		new MyMenuItem(m, layout.getMenuLabel(), layout, 0, 0);
		layout = new GridGroupVertically(this);
		new MyMenuItem(m, layout.getMenuLabel(), layout, 0, 0);
		m.addSeparator();
		layout = new SugiyamaLayout(this);
		new MyMenuItem(m, layout.getMenuLabel(), layout, 0, 0);
		mb.add(m);

		m = new JMenu("Options");
		viewModeMenuItem(m, this);
		m.addSeparator();
		bestEdgeMenuItem(m, this);
		m.addSeparator();
		showMenuItem(m, this);
		mb.add(m);
		m_optionMenu = m;

		m  = new JMenu("Other");
		m1 = new JMenu("Tab box");
		Do.tabMenuItem(m1, this);
		m.add(m1);
		m1 = new JMenu("Visible relations");
		Do.visibleMenuItem(m1, this);
		m.add(m1);
		m1 = new JMenu("Active relations");
		Do.activeMenuItem(m1, this);
		m.add(m1);
		mb.add(m);
		m1 = new JMenu("Navigate");
		Do.navigateEdgeMenu(m1, this);
		Do.navigateEntityMenu(m1, this);
		tocMenuItem(m1, this);
		m.add(m1);
		m1 = new JMenu("Query");
		Do.queryMenuItem(m1, this);
		m.add(m1);

		m1 = new JMenu("Elide");
		Do.hideMenuItem(m1, this);
		m.add(m1);

		mb.add(m);

		m = new JMenu("Help");
		Do.helpMenuItem(m, this, "Editor");
		mb.add(m);
		return mb;
	}

	// Overrides LandscapeEditorCore.genGUI() completely

	protected void genGUI(Dimension diagPercent) 
	{
		MsgOut.dprintln("App: gen GUI");

		genDropDowns();
		genMainGUI(diagPercent);
	}

	protected void setMenuCheck(int key, boolean state) 
	{
		JMenuItem			item;
		MyCheckBoxMenuItem	checkbox;

		for (int i = 0; i < m_optionMenu.getItemCount(); i++) {
			item = m_optionMenu.getItem(i);
			if (item != null && (item instanceof MyCheckBoxMenuItem)) {
				checkbox = (MyCheckBoxMenuItem) item;
				if (checkbox.getKey() == key) {
					checkbox.setSelected(state);
					return;
		}	}	}
	}

	protected boolean testForClose() 
	{
		if (!m_diagram.getChangedFlag())
			return true;

		System.out.print("\007");	// Bell

		int rc = JOptionPane.showConfirmDialog(m_frame, "Landscape has been changed.\nShould it be saved?", "Landscape Changed", JOptionPane.YES_NO_CANCEL_OPTION);

		switch(rc)
		{
		case JOptionPane.YES_OPTION:
			saveLs();
			// Drop through
		case JOptionPane.NO_OPTION:
			return true;
		}

		return false;
	}

	// Overrides LandscapeEditorCore.processKey() 

	public void processKey(int key, int modifiers, Object object) 
	{
		if (modifiers == Event.CTRL_MASK) {
			switch(key)	{
			case Do.QUIT_PROGRAM:
				if (testForClose()) {
					System.exit(0);
				}
				return;
			case Do.CLOSE:
				if (!testForClose()) {
					return;
				}
				if (openFrames == 1) {
					System.exit(0);
					return;
				}
				m_frame.setVisible(false);
				m_frame.dispose();
				openFrames--;
				return;
		}	}
		super.processKey(key, modifiers, object);
	}

	public void followLink(String file, int target) 
	{
//		System.out.println("followLink: " + target);

		if (target == LsLink.TARGET_NEW) {
			LandscapeEditorApp ap = new LandscapeEditorApp();

			ap.lsPath = file;
 
			LandscapeEditorFrame af = LandscapeEditorFrame.create(ap);
			openFrames++;
		} else if (target == LsLink.TARGET_APP) {
			super.followLink(file, target);
		} else {
			attach(file);
		}
	}

	public String getParameter(String name) 
	{
		return System.getProperty(name);
	}
	
/*
	// For debugging
	
	public void setBounds(int x, int y, int width, int height)
	{
		super.setBounds(x, y, width, height);
		System.out.println("LandscapeEditorApp.setBounds(" + getBounds() + ")");
	}
*/
}
